import asyncio
import os
import site
import socket
import subprocess

from fhempy.lib.generic import FhemModule

from .. import fhem


class esphome(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.proc = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        
        self._set_list = {"start": {}, "stop": {}, "restart": {}}
        await self.set_set_config(self._set_list)
        self._attr_list = {
            "disable": {"default": "0", "options": "0,1"},
            "port_dashboard": {"default": "6052", "help": "Default port ist 6052"},
        }
        await self.set_attr_config(self._attr_list)

        if self._attr_disable == "1":
            return

        await self.start_process()

        if await fhem.init_done(hash) == 1:
            # create weblinks on first define
            self.create_async_task(self.create_weblink())

    async def start_process(self):
        my_env = os.environ
        my_env["PATH"] = site.getuserbase() + "/bin:" + my_env["PATH"]
        self._esphomeargs = [
            site.getuserbase() + "/bin/esphome",
            "dashboard",
            "esphome_config/",
            "--port",
            self._attr_port_dashboard,
        ]

        try:
            self.proc = subprocess.Popen(self._esphomeargs, env=my_env)
        except Exception:
            self.logger.exception("Failed to execute esphome")
            try:
                self._esphomeargs = [
                    "esphome",
                    "dashboard",
                    "esphome_config/",
                    "--port",
                    self._attr_port_dashboard,
                ]
                self.proc = subprocess.Popen(self._esphomeargs)
            except Exception:
                self.logger.exception("Failed to execute esphome")
                return "Failed to execute esphome"

        await fhem.readingsSingleUpdate(self.hash, "state", "running", 1)

    async def stop_process(self):
        if self.proc:
            await fhem.readingsSingleUpdate(self.hash, "state", "stopping", 1)
            self.proc.kill()

            stop_tries = 0
            # give zigbee2mqtt some time to stop
            await asyncio.sleep(3)
            while self.proc.poll is None and stop_tries < 5:
                await asyncio.sleep(5)
                self.proc.terminate()
                stop_tries += 1

            if self.proc.poll is None:
                self.logger.error("Failed to stop esphome process")
                await fhem.readingsSingleUpdate(self.hash, "state", "failed to stop", 1)
            else:
                # this should never block, as poll says process finished already
                # this should prevent zombie processes
                self.proc.wait(0.1)
                self.proc = None
                await fhem.readingsSingleUpdate(self.hash, "state", "stopped", 1)
            self.proc = None

    async def create_weblink(self):
        if await fhem.checkIfDeviceExists(
            self.hash, "TYPE", "weblink", "NAME", "esphome_dashboard"
        ):
            return

        hostname = socket.gethostname()
        local_ip = socket.gethostbyname(hostname)
        await fhem.CommandDefine(
            self.hash, "esphome_dashboard weblink iframe http://" + local_ip + ":6052/"
        )
        await fhem.CommandAttr(
            self.hash,
            (
                "esphome_dashboard htmlattr width='900' height='700' "
                "frameborder='0' marginheight='0' marginwidth='0'"
            ),
        )
        await fhem.CommandAttr(self.hash, "esphome_dashboard room ESPHome")

    # FHEM FUNCTION
    async def Undefine(self, hash):
        await self.stop_process()
        return await super().Undefine(hash)

    async def set_attr_disable(self, hash):
        if self._attr_disable == "0":
            await self.start_process()
        else:
            await self.stop_process()

    async def set_start(self, hash, params):
        self.create_async_task(self._restart())

    async def set_stop(self, hash, params):
        self.create_async_task(self.stop_process())

    async def _restart(self):
        await self.stop_process()
        await self.start_process()

    async def set_restart(self, hash, params):
        self.create_async_task(self._restart())



"""Contains the Async Smartmeter API Client."""
import asyncio
import json
import logging
import socket
from datetime import datetime
from urllib import parse

import aiohttp
import async_timeout
import constants as const
from errors import SmartmeterLoginError
from lxml import html

logger = logging.getLogger(__name__)

TIMEOUT = 20


class AsyncSmartmeter:
    """Async Smartmeter Client."""

    def __init__(self, username, password, session=None, timeout=TIMEOUT):
        """Access the Smart Meter API asynchronously.

        Args:
            username (str): Username used for API Login
            password (str): Password used for API Login
            session (aiohttp.ClientSession): An optional session object
            timeout (int): Timeout for all session calls. Defaults to TIMEOUT.
        """
        self._username = username
        self._password = password
        self._session = session or aiohttp.ClientSession()
        self._timeout = timeout
        self._access_token = None
        self._api_gateway_token = None

    async def _get_login_action(self):
        login_url = const.AUTH_URL + "auth?" + parse.urlencode(const.LOGIN_ARGS)
        async with self._session.get(login_url) as response:
            tree = html.fromstring(await response.text())
            return tree.xpath("(//form/@action)")[0]

    async def _get_auth_code(self):

        action = await self._get_login_action()

        async with self._session.request(
            "POST",
            action,
            data={"username": self._username, "password": self._password},
            allow_redirects=False,
        ) as resp:
            if "Location" not in resp.headers:
                raise SmartmeterLoginError(
                    "Authentication failed. Check user credentials."
                )
            auth_code = resp.headers["Location"].split("&code=", 1)[1]
            return auth_code

    async def _get_api_key(self, access_token):
        async with self._session.request(
            "GET", const.PAGE_URL, data={"Authorization": f"Bearer {access_token}"}
        ) as result:
            tree = html.fromstring(await result.text())
            scripts = tree.xpath("(//script/@src)")
            for script in scripts:
                async with self._session.request(
                    "GET", const.PAGE_URL + script
                ) as response:
                    if const.MAIN_SCRIPT_REGEX.match(script):
                        for match in const.API_GATEWAY_TOKEN_REGEX.findall(
                            await response.text()
                        ):
                            return match
        return None

    async def refresh_token(self):
        """Create a valid access token."""
        async with self._session.request(
            "POST",
            const.AUTH_URL + "token",
            data=const.build_access_token_args(code=await self._get_auth_code()),
        ) as response:
            if response.status != 200:
                raise SmartmeterLoginError(
                    "Authentication failed. Check user credentials."
                )
            self._access_token = json.loads(await response.text())["access_token"]
            self._api_gateway_token = await self._get_api_key(self._access_token)

        logger.debug("Successfully authenticated Smart Meter API")

    async def async_get_access_token(self):
        """Return a valid access token."""
        pass

    def _dt_string(self, dt):
        return dt.strftime(const.API_DATE_FORMAT)[:-3] + "Z"

    async def _get_first_zaehlpunkt(self):
        """Get first zaehlpunkt."""
        zaehlpunkte = await self.get_zaehlpunkte()
        return zaehlpunkte[0]["zaehlpunkte"][0]["zaehlpunktnummer"]

    async def get_zaehlpunkte(self):
        """Get zaehlpunkte for currently logged in user."""
        return await self._request("zaehlpunkte")

    async def get_verbrauch_raw(
        self,
        date_from,
        date_to=None,
        zaehlpunkt=None,
    ):
        """Get verbrauch_raw from the API."""
        if date_to is None:
            date_to = datetime.now()
        if zaehlpunkt is None:
            zaehlpunkt = await self._get_first_zaehlpunkt()
        endpoint = f"messdaten/zaehlpunkt/{zaehlpunkt}/verbrauchRaw"
        query = {
            "dateFrom": self._dt_string(date_from),
            "dateTo": self._dt_string(date_to),
            "granularity": "DAY",
        }
        return await self._request(endpoint, query=query)

    async def get_verbrauch(
        self,
        date_from,
        date_to=None,
        zaehlpunkt=None,
    ):
        """Get verbrauch_raw from the API."""
        if date_to is None:
            date_to = datetime.now()
        if zaehlpunkt is None:
            zaehlpunkt = await self._get_first_zaehlpunkt()
        endpoint = f"messdaten/zaehlpunkt/{zaehlpunkt}/verbrauch"
        query = const.build_verbrauchs_args(
            dateFrom=self._dt_string(date_from), dateTo=self._dt_string(date_to)
        )
        return await self._request(endpoint, query=query)

    async def tages_verbrauch(
        self,
        day: datetime,
        zaehlpunkt=None,
    ):
        """Get verbrauch_raw from the API."""
        if zaehlpunkt is None:
            zaehlpunkt = await self._get_first_zaehlpunkt()
        endpoint = f"messdaten/zaehlpunkt/{zaehlpunkt}/verbrauch"
        query = const.build_verbrauchs_args(
            dateFrom=self._dt_string(day.replace(hour=0, minute=0, second=0))
        )
        return await self._request(endpoint, query=query)

    async def profil(self):
        """Get profil of logged in user."""
        return await self._request("user/profile", const.API_URL_ALT)

    async def zaehlpunkte(self):
        """Returns zaehlpunkte for currently logged in user."""
        return await self._request("zaehlpunkte")

    async def consumptions(self):
        """Returns response from 'consumptions' endpoint."""
        return await self._request("zaehlpunkt/consumptions")

    async def base_information(self):
        """Returns response from 'baseInformation' endpoint."""
        return await self._request("zaehlpunkt/baseInformation")

    async def meter_readings(self):
        """Returns response from 'meterReadings' endpoint."""
        return await self._request("zaehlpunkt/meterReadings")

    async def _request(
        self,
        endpoint,
        base_url=None,
        method="GET",
        data=None,
        query=None,
    ):
        """Send requests to the Smartmeter API."""
        if base_url is None:
            base_url = const.API_URL
        url = "{0}{1}".format(base_url, endpoint)

        if query:
            separator = "?" if "?" not in endpoint else "&"
            url += separator + parse.urlencode(query)

        logger.debug(f"REQUEST: {url}")

        headers = {"Authorization": f"Bearer {self._access_token}"}
        if self._api_gateway_token is not None:
            headers.update({"X-Gateway-APIKey": self._api_gateway_token})

        try:
            async with async_timeout.timeout(self._timeout):
                response = await self._session.request(
                    method, url, headers=headers, json=data
                )
                logger.debug(f"REQUEST: {response}")
                if response.status == 401:
                    await self.refresh_token()
                    return await self._request(endpoint, base_url, method, data, query)
                return await response.json()

        except asyncio.TimeoutError as exception:
            logger.error(f"Timeout error fetching information from {url} - {exception}")
        except (KeyError, TypeError) as exception:
            logger.error(f"Error parsing information from {url} - {exception}")
        except (aiohttp.ClientError, socket.gaierror) as exception:
            logger.error(f"Error fetching information from {url} - {exception}")
        except Exception as exception:
            logger.error(f"Something really wrong happened! - {exception}")



import asyncio

from fhempy.lib import fhem, fhem_pythonbinding
from fhempy.lib.generic import FhemModule
from meross_iot.controller.mixins.garage import GarageOpenerMixin
from meross_iot.controller.mixins.light import LightMixin
from meross_iot.controller.mixins.roller_shutter import RollerShutterTimerMixin
from meross_iot.controller.mixins.spray import SprayMixin
from meross_iot.controller.mixins.thermostat import ThermostatModeMixin
from meross_iot.controller.mixins.toggle import ToggleMixin, ToggleXMixin
from meross_iot.model.enums import Namespace, OnlineStatus, SprayMode, ThermostatMode


class meross_device:
    def __init__(self, logger, fhemdevice: FhemModule):
        self.logger = logger
        self.fhemdev = fhemdevice
        self.hash = fhemdevice.hash

    async def Define(self, hash, args, argsh):
        self._setupdev_name = args[3]
        self._deviceid = args[4]
        self._device = None
        self.hash["DEVICEID"] = self._deviceid

        self._setupdev = None
        await fhem.readingsSingleUpdate(self.hash, "state", "ready", 1)
        self.fhemdev.create_async_task(self._init_device())

    async def _get_set_commands(self):
        set_conf = {}

        if isinstance(self._device, ToggleXMixin) or isinstance(
            self._device, ToggleMixin
        ):
            for channel in self._device.channels:
                if channel.is_master_channel:
                    set_conf["on"] = {
                        "function": "set_on",
                        "function_param": channel.index,
                    }
                    set_conf["off"] = {
                        "function": "set_off",
                        "function_param": channel.index,
                    }
                    set_conf["toggle"] = {
                        "function": "set_toggle",
                        "function_param": channel.index,
                    }
                else:
                    set_conf[f"on_{channel.index}"] = {
                        "function": "set_on",
                        "function_param": channel.index,
                    }
                    set_conf[f"off_{channel.index}"] = {
                        "function": "set_off",
                        "function_param": channel.index,
                    }
                    set_conf[f"toggle_{channel.index}"] = {
                        "function": "set_toggle",
                        "function_param": channel.index,
                    }

        if isinstance(self._device, RollerShutterTimerMixin):
            set_conf["open"] = {}
            set_conf["close"] = {}
            set_conf["stop"] = {}

        if isinstance(self._device, GarageOpenerMixin):
            set_conf["open"] = {}
            set_conf["close"] = {}

        if isinstance(self._device, ThermostatModeMixin):
            set_conf["mode"] = {
                "args": ["value"],
                "options": "heat,cool,economy,auto,manual",
                "function": "set_thermostat_mod",
            }
            set_conf["desiredTemp"] = {
                "args": ["value"],
                "options": "slider,5,0.5,30,1",
                "format": "float",
            }
            set_conf["on"] = {"function": "set_thermostat_on"}
            set_conf["off"] = {"function": "set_thermostat_off"}

        if isinstance(self._device, LightMixin):
            if self._device.get_supports_rgb():
                set_conf["rgb"] = {"args": ["value"], "options": "colorpicker,RGB"}
            if self._device.get_supports_luminance():
                set_conf["brightness"] = {
                    "args": ["value"],
                    "options": "colorpicker,BRI,0,1,100",
                }
            if self._device.get_supports_temperature():
                set_conf["ct"] = {
                    "args": ["value"],
                    "options": "colorpicker,CT,2000,1,6500",
                }

        if isinstance(self._device, SprayMixin):
            set_conf["intermittent"] = {}
            set_conf["off"] = {
                "function": "set_off",
                "function_param": 0,
            }
            set_conf["continuous"] = {}

        await self.fhemdev.set_set_config(set_conf)

    async def set_thermostat_mode(self, hash, params):
        new_mode = ThermostatMode(params["value"].upper())
        await self._device.async_set_thermostat_config(mode=new_mode)

    async def set_thermostat_on(self, hash, params):
        await self._device.async_set_thermostat_config(on_not_off=True)

    async def set_thermostat_off(self, hash, params):
        await self._device.async_set_thermostat_config(on_not_off=False)

    async def set_desiredTemp(self, hash, params):
        await self._device.async_set_thermostat_config(
            manual_temperature_celsius=params["value"]
        )

    async def set_intermittent(self, hash, params):
        await self._device.async_set_mode(SprayMode.INTERMITTENT)

    async def set_continuous(self, hash, params):
        await self._device.async_set_mode(SprayMode.CONTINUOUS)

    async def set_rgb(self, hash, params):
        rgb = params["value"]
        red = int(rgb[0:2], base=16)
        green = int(rgb[2:4], base=16)
        blue = int(rgb[4:6], base=16)
        await self._device.async_set_light_color(rgb=(red, green, blue))

    async def set_brightness(self, hash, params):
        bri = params["value"]
        await self._device.async_set_light_color(luminance=bri)

    async def set_ct(self, hash, params):
        ct = params["value"]
        await self._device.async_set_light_color(temperature=ct)

    async def set_on(self, hash, params):
        await self._device.async_turn_on(params["function_param"])

    async def set_off(self, hash, params):
        if isinstance(self._device, SprayMixin):
            await self._device.async_set_mode(SprayMode.OFF)
        else:
            await self._device.async_turn_off(params["function_param"])

    async def set_toggle(self, hash, params):
        await self._device.async_toggle(params["function_param"])

    async def set_open(self, hash, params):
        await self._device.async_open()

    async def set_close(self, hash, params):
        await self._device.async_close()

    async def set_stop(self, hash, params):
        await self._device.async_stop()

    async def _init_device(self):
        try:
            await self._connect_to_setup_device()
            await self._setup_device()
            await self._get_set_commands()
            await self.update_readings()
        except Exception as ex:
            self.logger.exception(ex)

    async def _connect_to_setup_device(self):
        while self._setupdev is None or self._setupdev.ready is False:
            await asyncio.sleep(1)
            self._setupdev = fhem_pythonbinding.getFhemPyDeviceByName(
                self._setupdev_name
            )
            if self._setupdev:
                self._setupdev = self._setupdev.meross_device

    async def _async_push_notification_received(
        self, namespace: Namespace, data: dict, device_internal_id: str
    ):
        update_state = False
        full_update = False

        if namespace == Namespace.CONTROL_UNBIND:
            self.logger.warning(
                f"Received unbind event. Removing device {self._device.name} from FHEM"
            )
            await self.platform.async_remove_entity(self.entity_id)
        elif namespace == Namespace.SYSTEM_ONLINE:
            self.logger.warning(f"Device {self._device.name} reported online event.")
            online = OnlineStatus(int(data.get("online").get("status")))
            update_state = True
            full_update = online == OnlineStatus.ONLINE

        elif namespace == Namespace.HUB_ONLINE:
            self.logger.warning(
                f"Device {self._device.name} reported (HUB) online event."
            )
            online = OnlineStatus(int(data.get("status")))
            update_state = True
            full_update = online == OnlineStatus.ONLINE
        else:
            update_state = True
            full_update = False

        if full_update:
            await self._device.async_update()

        if update_state:
            await self.update_readings()

    async def _setup_device(self):
        self._device = self._setupdev.get_device_by_id(self._deviceid)

        if self._device is not None:
            await self._device.async_update()
            self._device.register_push_notification_handler_coroutine(
                self._async_push_notification_received
            )

    async def update_readings(self):
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "uuid", self._device.uuid)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "name", self._device.name)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "firmware_version", self._device.firmware_version
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "hardware_version", self._device.hardware_version
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "internal_id", self._device.internal_id
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "mqtt_host", self._device.mqtt_host
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "mqtt_port", self._device.mqtt_port
        )
        await fhem.readingsBulkUpdateIfChanged(self.hash, "type", self._device.type)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "online_status", self._device.online_status.name
        )

        state_val = "unknown"
        if isinstance(self._device, ToggleXMixin) or isinstance(
            self._device, ToggleMixin
        ):
            for channel in self._device.channels:
                if channel.is_master_channel:
                    state_val = "off"
                    if self._device.is_on(channel.index):
                        state_val = "on"
                else:
                    state = "on" if self._device.is_on(channel.index) else "off"
                    await fhem.readingsBulkUpdateIfChanged(
                        self.hash, f"switch_{channel.index}", state
                    )

        if isinstance(self._device, GarageOpenerMixin):
            state_val = "closed"
            if self._device.get_is_open():
                state_val = "open"

        if isinstance(self._device, RollerShutterTimerMixin):
            status = self._device.get_status()
            state_val = "unknown"
            if status == 1:
                state_val = "open"
            elif status == 2:
                state_val = "closed"
            elif status == 0:
                # stopped, keep last state instead of stopped
                # state_val = "stopped"
                pass

            pct = self._device.get_position()
            await fhem.readingsBulkUpdateIfChanged(self.hash, "pct", pct)

        if isinstance(self._device, LightMixin):
            ct = self._device.get_color_temperature()
            await fhem.readingsBulkUpdateIfChanged(self.hash, "ct", ct)

            bri = self._device.get_luminance()
            await fhem.readingsBulkUpdateIfChanged(self.hash, "brightness", bri)

            rgb_tuple = self._device.get_rgb_color()
            rgb = f"{rgb_tuple[0]:02x}{rgb_tuple[1]:02x}{rgb_tuple[2]:02x}"
            await fhem.readingsBulkUpdateIfChanged(self.hash, "rgb", rgb)

        if isinstance(self._device, ThermostatModeMixin):
            thermostat_state = self._device.get_thermostat_state()
            state_val = "on" if thermostat_state.is_on else "off"
            mode = thermostat_state.mode
            if mode:
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, "mode", mode.name.lower()
                )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "desiredTemp", thermostat_state.target_temperature_celsius
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "min_temp", thermostat_state.min_temperature_celsius
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "max_temp", thermostat_state.max_temperature_celsius
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "heat_temp", thermostat_state.heat_temperature_celsius
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "cool_temp", thermostat_state.cool_temperature_celsius
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "eco_temp", thermostat_state.eco_temperature_celsius
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "manual_temp", thermostat_state.manual_temperature_celsius
            )

        if isinstance(self._device, SprayMixin):
            mode = self._device.get_current_mode()
            if mode == SprayMode.CONTINUOUS:
                state_val = "continuous"
            elif mode == SprayMode.INTERMITTENT:
                state_val = "intermittent"
            elif mode == SprayMode.OFF:
                state_val = "off"

        await fhem.readingsBulkUpdateIfChanged(self.hash, "state", state_val)
        await fhem.readingsEndUpdate(self.hash, 1)



import asyncio

from api.client import AsyncSmartmeter
from fhempy.lib.generic import FhemModule

from .. import fhem, utils


class wienernetze_smartmeter(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        set_list_conf = {"update": {}}
        await self.set_set_config(set_list_conf)
        if len(args) != 5:
            return (
                "Usage: define devname fhempy wienernetze_smartmeter USERNAME PASSWORD"
            )
        self.username = args[3]
        self.password = args[4]

        self._updateloop = self.create_async_task(self.update_loop())

    async def update_loop(self):
        client = AsyncSmartmeter(self.username, self.password)
        await client.refresh_token()
        data = await client.base_information()
        await self.update_readings(data)
        while True:
            data = await client.consumptions()
            await self.update_readings(data)
            data = await client.meter_readings()
            await self.update_readings(data)
            await asyncio.sleep(3600)

    async def update_readings(self, data):
        flat_data = utils.flatten_json(data)
        await fhem.readingsBeginUpdate(self.hash)
        for name in flat_data:
            await fhem.readingsBulkUpdateIfChanged(self.hash, name, flat_data[name])
        await fhem.readingsEndUpdate(self.hash, 1)



from fhempy.lib import fhem
from fhempy.lib.ble_monitor.blemonitor import BLEmonitor
from fhempy.lib.ble_monitor.bt_helpers import BT_INTERFACES

from .. import generic


class ble_monitor(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.blemonitor = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        attr_config = {
            "hci_interface": {
                "default": "0",
                "options": ",".join(list(map(str, BT_INTERFACES))),
                "format": "int",
                "help": f"{BT_INTERFACES}",
                "function": "set_attr_generic",
            },
            "encryption_key": {
                "default": "",
                "help": (
                    "Get encryption key (BLE KEY) "
                    "via https://github.com/PiotrMachowski/Xiaomi-cloud-tokens-extractor"
                ),
                "function": "set_attr_generic",
            },
        }
        await self.set_attr_config(attr_config)

        if len(args) != 4:
            return "Usage: define my_miscale fhempy ble_monitor 11:22:33:44:55:66"

        self.mac_addr = args[3]

        self.blemonitor = BLEmonitor.getInstance(self.logger)
        self.blemonitor.register_device(self)

    def mac(self):
        return self.mac_addr

    def hci(self):
        return self._attr_hci_interface

    def encryption_key(self):
        if self._attr_encryption_key == "":
            return None
        return self._attr_encryption_key

    async def received_data(self, data):
        try:
            await fhem.readingsBeginUpdate(self.hash)
            for reading in data:
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, reading, data[reading]
                )
            await fhem.readingsEndUpdate(self.hash, 1)
        except Exception:
            self.logger.exception("Failed to update readings")

    async def set_attr_generic(self, hash):
        self.blemonitor.unregister_device(self)
        self.blemonitor.register_device(self)

    async def Undefine(self, hash):
        if self.blemonitor:
            self.blemonitor.unregister_device(self)
        return await super().Undefine(hash)
